/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@gnome.org>
 *
 *  Copyright 2003-2006 Iain Holmes
 *
 *  This file is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU Library General Public
 *  License as published by the Free Software Foundation;
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *
 */

#include <glib/gi18n.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkstock.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtklabel.h>

#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>

#include <gst/gst.h>

#include <marlin/marlin-stock.h>
#include <marlin/marlin-file-chooser.h>
#include <marlin/marlin-x-utils.h>
#include <marlin/marlin-utils.h>
#include <marlin/marlin-a11y-helper.h>

struct _ChooserOpenData {
	GtkWidget *chooser;
	
	GtkWidget *options_frame;
	GtkWidget *info_frame;

	/* Info widgets */
	GtkWidget *length;
	GtkWidget *sample_rate;
	GtkWidget *name;
	GtkWidget *channels;
	GtkWidget *mimetype;
	
	/* Transport controls */
	GtkWidget *play;
	GtkWidget *stop;

	/* Filesel hooks */
	GtkTreeSelection *selection;
	guint32 change_id;
	
	/* Preview */
	GstElement *player;
	GstElement *md_pipeline;
	GstPad *decode_pad;
};

static void
free_open_data (GtkFileChooser          *fc,
		struct _ChooserOpenData *od)
{
	if (od->player != NULL) {
		gst_element_set_state (od->player, GST_STATE_NULL);
		g_object_unref (G_OBJECT (od->player));
	}
	if (od->md_pipeline != NULL) {
		gst_element_set_state (od->md_pipeline, GST_STATE_NULL);
		g_object_unref (G_OBJECT (od->md_pipeline));
	}

	g_free (od);
}

static void G_GNUC_UNUSED
clear_info (struct _ChooserOpenData *od)
{
	char *markup;

	markup = g_strdup_printf ("<span weight=\"bold\">%s</span>",
				  _("Unknown"));
	gtk_label_set_markup (GTK_LABEL (od->name), markup);
	g_free (markup);

	gtk_label_set_text (GTK_LABEL (od->sample_rate), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->channels), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->length), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->mimetype), _("Unknown"));
}

/* Constructs a string out of a taglist: If there are multiple entries for a
   tag it concatenates them together as "a, b, c" */
static char *
get_tag (GstTagList *tags,
	 const char *tag)
{
	int i, count;
	GString *builder;
	char *string;

	builder = g_string_new ("");
	
	count = gst_tag_list_get_tag_size (tags, tag);
	if (count == 0) {
		g_string_free (builder, TRUE);
		return NULL;
	}

	for (i = 0; i < count; i++) {
		char *str;

		if (i != 0) {
			g_string_append (builder, ", ");
		}

		if (gst_tag_get_type (tag) == G_TYPE_STRING) {
			if (gst_tag_list_get_string_index (tags, tag, 
							   i, &str)) {
				g_string_append (builder, str);
				g_free (str);
			}
		} else {
			str = g_strdup_value_contents
				(gst_tag_list_get_value_index (tags, tag, i));
			g_string_append (builder, str);
			g_free (str);
		}
	}

	string = builder->str;
	g_string_free (builder, FALSE);

	return string;
}
				
static void
set_info (struct _ChooserOpenData *od,
	  GstTagList              *tags)
{
	char *str;

	str = get_tag (tags, GST_TAG_TITLE);
	if (str) {
		char *title;

		title = g_strdup_printf ("<span weight=\"bold\">%s</span>", 
					 str);
		g_free (str);
		gtk_label_set_markup (GTK_LABEL (od->name), title);
		g_free (title);
	} else {
		char *fullpath, *filename, *title;

		fullpath = gtk_file_chooser_get_preview_filename (GTK_FILE_CHOOSER (od->chooser));
		filename = g_path_get_basename (fullpath);
		g_free (fullpath);

		title = g_strdup_printf ("<span weight=\"bold\">%s</span>", 
					 filename);
		g_free (filename);
		gtk_label_set_markup (GTK_LABEL (od->name), title);
		g_free (title);
	}
}

static gboolean
md_message_cb (GstBus     *bus,
	       GstMessage *message,
	       gpointer    data)
{
	struct _ChooserOpenData *od = data;
	GstTagList *tags;
	
	switch (GST_MESSAGE_TYPE (message)) {
	case GST_MESSAGE_EOS:
		clear_info (od);
		break;

	case GST_MESSAGE_ERROR:
		clear_info (od);
		break;

	case GST_MESSAGE_TAG:
		gtk_file_chooser_set_preview_widget_active (GTK_FILE_CHOOSER (od->chooser), TRUE);
		gst_message_parse_tag (message, &tags);
		set_info (od, tags);
		gst_tag_list_free (tags);
		break;

	default:
		break;
	}

	return TRUE;
}

static void
unknown_type_cb (GstElement              *decodebin,
		 GstPad                  *pad,
		 GstCaps                 *caps,
		 struct _ChooserOpenData *od)
{
	gtk_file_chooser_set_preview_widget_active (GTK_FILE_CHOOSER (od->chooser), FALSE);
}

static void
have_type_cb (GstElement              *typefind,
	      guint                    probability,
	      GstCaps                 *caps,
	      struct _ChooserOpenData *od)
{
	char *type;
	const char *name;

	type = gst_caps_to_string (caps);

	name = gnome_vfs_mime_get_description (type);
	if (name) {
		gtk_label_set_text (GTK_LABEL (od->mimetype), name);
	} else {
		gtk_label_set_text (GTK_LABEL (od->mimetype), type);
	}
	g_free (type);
}

static void
new_pad (GstElement              *decodebin,
	 GstPad                  *pad,
	 gboolean                 final,
	 struct _ChooserOpenData *od)
{
	od->decode_pad = pad;
}

static void 
get_stream_info (struct _ChooserOpenData *od)
{
	GstCaps *caps;
	GstStructure *st;
	GstQuery *query;
	int value;

	if (od->decode_pad == NULL) {
		return;
	} 

	caps = gst_pad_get_caps (od->decode_pad);
	st = gst_caps_get_structure (caps, 0);

	if (gst_structure_get_int (st, "rate", &value)) {
		char *s = g_strdup_printf ("%d hz", value);
		gtk_label_set_text (GTK_LABEL (od->sample_rate), s);
		g_free (s);
	}

	if (gst_structure_get_int (st, "channels", &value)) {
		char *s = marlin_channels_to_string (value);
		gtk_label_set_text (GTK_LABEL (od->channels), s);
		g_free (s);
	}

	gst_caps_unref (caps);

	query = gst_query_new_duration (GST_FORMAT_TIME);
	if (gst_pad_query (od->decode_pad, query)) {
		gint64 duration;
		char *time_str;

		gst_query_parse_duration (query, NULL, &duration);
		time_str = marlin_ms_to_pretty_time (duration / 1000000, FALSE);
		gtk_label_set_text (GTK_LABEL (od->length), time_str);
		g_free (time_str);
	}

	gst_query_unref (query);
}

static void
get_metadata (struct _ChooserOpenData *od,
	      const char              *filename)
{
	GstElement *decodebin, *source;
	GstObject *typefind;
	GstStateChangeReturn sret;
	GstState state;

	od->md_pipeline = gst_pipeline_new (NULL);
	gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE (od->md_pipeline)),
			   md_message_cb, od);

	source = gst_element_factory_make ("gnomevfssrc", "source");
	decodebin = gst_element_factory_make ("decodebin", "decodebin");
	g_signal_connect (G_OBJECT (decodebin), "unknown-type",
			  G_CALLBACK (unknown_type_cb), od);
	g_signal_connect (G_OBJECT (decodebin), "new-decoded-pad",
			  G_CALLBACK (new_pad), od);

	typefind = gst_child_proxy_get_child_by_name (GST_CHILD_PROXY (decodebin),
						      "typefind");
	g_signal_connect (G_OBJECT (typefind), "have-type",
			  G_CALLBACK (have_type_cb), od);

	gst_bin_add_many (GST_BIN (od->md_pipeline), source, decodebin, NULL);
	gst_element_link (source, decodebin);
	
	g_object_set (G_OBJECT (source),
		      "location", filename,
		      NULL);
	
	sret = gst_element_set_state (od->md_pipeline, GST_STATE_PAUSED);
	
	if (sret == GST_STATE_CHANGE_ASYNC) {
		if (gst_element_get_state (od->md_pipeline, &state, NULL,
					   5 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
			clear_info (od);
		} else {
			/* Need to wait til the state has gone to PAUSED 
			   before we get the stream info */
			get_stream_info (od);
		}
	} else if (sret != GST_STATE_CHANGE_SUCCESS) {
		clear_info (od);
	}
}

static void
selection_changed (GtkFileChooser          *chooser,
		   struct _ChooserOpenData *od)
{
	char *filename = gtk_file_chooser_get_preview_filename (chooser);

	/* Destroy the old pipeline */
	if (od->md_pipeline != NULL) {
		gst_element_set_state (od->md_pipeline, GST_STATE_NULL);
		g_object_unref (G_OBJECT (od->md_pipeline));
		od->md_pipeline = NULL;
	}

	if (g_file_test (filename, G_FILE_TEST_IS_DIR)) {
		gtk_file_chooser_set_preview_widget_active (GTK_FILE_CHOOSER (od->chooser), FALSE);
		g_free (filename);
		return;
	}

	if (filename != NULL) {
		gtk_widget_set_sensitive (od->play, TRUE);
		get_metadata (od, filename);
		g_free (filename);
	} else {
		gtk_widget_set_sensitive (od->play, FALSE);
		clear_info (od);
	}
}

static gboolean
bus_message_cb (GstBus     *bus,
		GstMessage *message,
		gpointer    data)
{
	struct _ChooserOpenData *od = data;

	switch (GST_MESSAGE_TYPE (message)) {
	case GST_MESSAGE_EOS:
		gtk_widget_show (od->play);
		gtk_widget_hide (od->stop);
		break;

	default:
		break;
	}

	return TRUE;
}

static void
start_playing (GtkWidget               *button,
	       struct _ChooserOpenData *od)
{
	char *uri;

	/* Tear it down */
	if (od->player != NULL) {
		gst_element_set_state (od->player, GST_STATE_NULL);
		g_object_unref (G_OBJECT (od->player));
		od->player = NULL;
	}
	
	/* Build it up */
 	uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (od->chooser));
	
	od->player = gst_element_factory_make ("playbin", "player");
	gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE (od->player)),
			   bus_message_cb, od);

	g_object_set (G_OBJECT (od->player),
		      "uri", uri,
		      NULL);
	
	gst_element_set_state (od->player, GST_STATE_PLAYING);

	gtk_widget_hide (od->play);
	gtk_widget_show (od->stop);
}

static void
stop_playing (GtkWidget               *button,
	      struct _ChooserOpenData *od)
{
	gst_element_set_state (od->player, GST_STATE_READY);
	
	gtk_widget_hide (od->stop);
	gtk_widget_show (od->play);
}

static void
build_info_contents (struct _ChooserOpenData *od)
{
	GtkWidget *table, *label, *vbox, *spacer, *hbox;

	vbox = gtk_vbox_new (FALSE, 6);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (od->chooser), vbox);

	od->name = marlin_make_title_label (_("Untitled"));
	gtk_box_pack_start (GTK_BOX (vbox), od->name, FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	spacer = gtk_label_new ("    ");
	gtk_box_pack_start (GTK_BOX (hbox), spacer, FALSE, FALSE, 0);
	
	table = marlin_make_table (4, 2, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox), table, TRUE, TRUE, 0);

#if 0
	label = marlin_make_title_label (_("Name:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 0, GTK_FILL);

	gtk_label_set_line_wrap (GTK_LABEL (od->name), TRUE);
	PACK (table, od->name, 1, 0, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->name, ATK_RELATION_LABELLED_BY);
#endif

	label = marlin_make_title_label (_("Mime Type:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 0, GTK_FILL);

	od->mimetype = marlin_make_info_label (_("Unknown"));
	PACK (table, od->mimetype, 1, 0, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->mimetype, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Length:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 1, GTK_FILL);

	od->length = marlin_make_info_label (_("Unknown"));
	PACK (table, od->length, 1, 1, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->length, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Sample Rate:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 2, GTK_FILL);

	od->sample_rate = marlin_make_info_label (_("Unknown"));
	PACK (table, od->sample_rate, 1, 2, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->sample_rate, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Channels:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 3, GTK_FILL);

	od->channels = marlin_make_info_label (_("Unknown"));
	PACK (table, od->channels, 1, 3, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->channels, ATK_RELATION_LABELLED_BY);
	
	hbox = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	od->play = marlin_make_button (_("Preview File"), MARLIN_STOCK_PLAY);
	gtk_widget_set_sensitive (od->play, FALSE);
	g_signal_connect (G_OBJECT (od->play), "clicked",
			  G_CALLBACK (start_playing), od);
	gtk_box_pack_start (GTK_BOX (hbox), od->play, FALSE, FALSE, 0);

	od->stop = marlin_make_button (_("Stop Preview"), MARLIN_STOCK_STOP);
	g_signal_connect (G_OBJECT (od->stop), "clicked",
			  G_CALLBACK (stop_playing), od);
	gtk_box_pack_start (GTK_BOX (hbox), od->stop, FALSE, FALSE, 0);
	
	gtk_widget_show_all (vbox);
	gtk_widget_hide (od->stop);
}

GtkWidget *
marlin_file_open_dialog_new (GtkWindow *parent_window)
{
	struct _ChooserOpenData *od;
	GtkWidget *widget;
	GtkFileFilter *filter;
	char *filename;

	od = g_new0 (struct _ChooserOpenData, 1);
	
	widget = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
			       "action", GTK_FILE_CHOOSER_ACTION_OPEN,
			       "title", _("Open File"),
			       "local-only", FALSE,
			       "use-preview-label", FALSE,
			       NULL);
	od->chooser = widget;

	build_info_contents (od);

	/* Check if we can display the first selected file */
	filename = gtk_file_chooser_get_preview_filename (GTK_FILE_CHOOSER (widget));
 	get_metadata (od, filename);
	g_free (filename);
	
	g_signal_connect (G_OBJECT (od->chooser), "destroy",
			  G_CALLBACK (free_open_data), od);
	g_signal_connect (G_OBJECT (od->chooser), "update-preview",
			  G_CALLBACK (selection_changed), od);
	
	gtk_dialog_add_buttons (GTK_DIALOG (widget),
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_OK,
				NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (widget), GTK_RESPONSE_OK);

	filter = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter, _("All Files"));
	gtk_file_filter_add_pattern (filter, "*");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);

	return widget;
}

GtkWidget *
marlin_file_save_dialog_new (GtkWindow *parent_window)
{
	GtkWidget *widget;

	widget = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
			       "action", GTK_FILE_CHOOSER_ACTION_SAVE,
			       "title", _("Save Sample As..."),
			       NULL);
	gtk_dialog_add_buttons (GTK_DIALOG (widget),
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_SAVE, GTK_RESPONSE_OK,
				NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (widget), GTK_RESPONSE_OK);
	gtk_window_set_default_size (GTK_WINDOW (widget), 600, 400);

	return widget;
}
