/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@gnome.org>
 *
 *  Copyright 2002 - 2007 Iain Holmes
 *
 *  This file is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public
 *  License as published by the Free Software Foundation;
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>

#include <samplerate.h>

#include <marlin/marlin-block.h>
#include <marlin/marlin-channel.h>
#include <marlin/marlin-marshal.h>
#include <marlin/marlin-sample.h>
#include <marlin/marlin-jack.h>

#define MARLIN_JACK_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), MARLIN_JACK_TYPE, MarlinJackPrivate))

typedef enum {
	MARLIN_JACK_MODE_NOTHING,
	MARLIN_JACK_MODE_STOPPED,
	MARLIN_JACK_MODE_PREROLLING,
	MARLIN_JACK_MODE_PLAYING,
	MARLIN_JACK_MODE_RECORDING
} MarlinJackMode;

enum {
	PROP_0,
	PROP_SAMPLE
};

enum {
	EOS,
	STATUS_CHANGED,
	POSITION_CHANGED,
	LAST_SIGNAL
};

#define DEFAULT_RB_SIZE 65536
const size_t frame_size = sizeof (jack_default_audio_sample_t);

struct _port_data {
	MarlinJack *jack;

	MarlinChannel *channel; /* The channel associated with the port */
	MarlinBlock *current_block; /* A pointer to the current block */
	guint64 offset; /* Where we are in the channel */
	guint64 offset_in_block; /* Where we are inside current_block */

	jack_port_t *playback; /* The playback port */
	jack_port_t *record; /* The record port */

	jack_ringbuffer_t *rb;
	float *rb_buffer, *tmp_buffer;

	SRC_STATE *state;

	gboolean writer_finished;
};

struct _MarlinJackPrivate {
	jack_client_t *client;
	jack_nframes_t bufsize;
	jack_default_audio_sample_t *jack_buf;

	MarlinSample *sample;
	gulong sample_notify_id;

	MarlinJackMode mode;

	guint channels;

	guint rate, jrate;
	double src_ratio;

	struct _port_data **port_data;

	guint writer_id;

	gboolean eos;
	guint64 position;

	guint64 start;
	guint64 finish;
	gboolean repeating;
};

static GObjectClass *parent_class;
static guint32 signals[LAST_SIGNAL] = { 0, };

static long process_more_frames (gpointer userdata, float **data);

GQuark
marlin_jack_error_quark (void)
{
	static GQuark quark = 0;
	if (quark == 0) {
		quark = g_quark_from_static_string ("marlin-jack-error-quark");
	}

	return quark;
}

static void
free_ports (MarlinJack *jack)
{
	MarlinJackPrivate *priv;
	int i;

	priv = jack->priv;

	for (i = 0; i < priv->channels; i++) {
		struct _port_data *pd = priv->port_data[i];

		jack_port_unregister (priv->client, pd->playback);
		jack_port_unregister (priv->client, pd->record);

		if (pd->rb) {
			jack_ringbuffer_free (pd->rb);
		}
		g_free (pd->rb_buffer);
		g_free (pd->tmp_buffer);

		if (pd->state) {
			src_delete (pd->state);
		}
		g_free (pd);
	}
	g_free (priv->port_data);
}

static void
setup_ports (MarlinJack *jack)
{
	MarlinJackPrivate *priv;
	int i;

	priv = jack->priv;

	/* We always want to set up two ports, even for mono samples */
	priv->port_data = g_new0 (struct _port_data *, 2);
	for (i = 0; i < 2; i++) {
		struct _port_data *pd;
		char *name;

		pd = g_new0 (struct _port_data, 1);

		pd->jack = jack;
		/* Create the ports */
		name = g_strdup_printf ("playback%d", i + 1);
		pd->playback = jack_port_register (priv->client, name,
						   JACK_DEFAULT_AUDIO_TYPE,
						   JackPortIsOutput, 0);
                g_free (name);
		
		name = g_strdup_printf ("record%d", i + 1);
		pd->record = jack_port_register (priv->client, name,
						 JACK_DEFAULT_AUDIO_TYPE,
						 JackPortIsInput, 0);
                g_free (name);

		priv->port_data[i] = pd;
	}
}

static void
finalize (GObject *object)
{
	MarlinJack *jack;
	MarlinJackPrivate *priv;

	jack = (MarlinJack *) object;
	priv = jack->priv;

	jack_client_close (priv->client);

	parent_class->finalize (object);
}

static void
dispose (GObject *object)
{
	MarlinJack *jack;
	MarlinJackPrivate *priv;

	jack = (MarlinJack *) object;
	priv = jack->priv;

	if (priv->writer_id > 0) {
		g_source_remove (priv->writer_id);
		priv->writer_id = 0;
	}

	if (priv->sample) {
		g_object_unref (priv->sample);
		priv->sample = NULL;
	}

	parent_class->dispose (object);
}

static double
calculate_rate_ratio (guint  src_rate,
		      guint dest_rate)
{
	double ratio;

	if (src_rate == 0) {
		return 1.0;
	}

	ratio = (1.0 * dest_rate) / src_rate;
	if (src_is_valid_ratio (ratio) == FALSE) {
		ratio = 1.0;
	}

	return ratio;
}

static void
sample_changed (GObject    *object,
		GParamSpec *pspec,
		MarlinJack *jack)
{
	MarlinJackPrivate *priv = jack->priv;
	
	if (strcmp (pspec->name, "sample_rate") == 0) {
		g_object_get (object,
			      "sample_rate", &priv->rate,
			      NULL);
		
		priv->src_ratio = calculate_rate_ratio (priv->rate, 
							priv->jrate);
		return;
	}

	if (strcmp (pspec->name, "channels") == 0) {
		free_ports (jack);
		g_object_get (object,
			      "channels", &priv->channels,
			      NULL);
		setup_ports (jack);
		return;
	}
}

static void
set_property (GObject      *object,
	      guint         prop_id,
	      const GValue *value,
	      GParamSpec   *pspec)
{
	MarlinJack *jack;
	MarlinJackPrivate *priv;

	jack = (MarlinJack *) object;
	priv = jack->priv;

	switch (prop_id) {
	case PROP_SAMPLE:
		if (priv->sample) {
			free_ports (jack);

			g_signal_handler_disconnect (priv->sample, 
						     priv->sample_notify_id);
			g_object_unref (priv->sample);
		}

		priv->sample = (MarlinSample *) g_value_dup_object (value);
		g_object_get (G_OBJECT (priv->sample),
			      "channels", &priv->channels,
			      "sample_rate", &priv->rate,
			      NULL);
		
		priv->sample_notify_id = g_signal_connect (priv->sample,
							   "notify",
							   G_CALLBACK (sample_changed),
							   jack);

		setup_ports (jack);
		break;

	default:
		break;
	}
}

static void
get_property (GObject    *object,
	      guint       prop_id,
	      GValue     *value,
	      GParamSpec *pspec)
{
}

static void
class_init (MarlinJackClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->finalize = finalize;
	o_class->dispose = dispose;
	o_class->set_property = set_property;
	o_class->get_property = get_property;

	g_type_class_add_private (klass, sizeof (MarlinJackPrivate));
	parent_class = g_type_class_peek_parent (klass);

	g_object_class_install_property (o_class,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample",
							      "", "",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_WRITABLE |
							      G_PARAM_STATIC_NAME |
							      G_PARAM_STATIC_NICK |
							      G_PARAM_STATIC_BLURB));

	signals[EOS] = g_signal_new ("eos",
				     G_TYPE_FROM_CLASS (klass),
				     G_SIGNAL_RUN_FIRST |
				     G_SIGNAL_NO_RECURSE,
				     G_STRUCT_OFFSET (MarlinJackClass, eos),
				     NULL, NULL,
				     g_cclosure_marshal_VOID__VOID,
				     G_TYPE_NONE, 0);
	signals[STATUS_CHANGED] = g_signal_new ("status-changed",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST |
						G_SIGNAL_NO_RECURSE,
						G_STRUCT_OFFSET (MarlinJackClass, status_changed),
						NULL, NULL,
						g_cclosure_marshal_VOID__INT,
						G_TYPE_NONE,
						1, G_TYPE_INT);
	signals[POSITION_CHANGED] = g_signal_new ("position-changed",
						  G_TYPE_FROM_CLASS (klass),
						  G_SIGNAL_RUN_FIRST |
						  G_SIGNAL_NO_RECURSE,
						  G_STRUCT_OFFSET (MarlinJackClass, position_changed),
						  NULL, NULL,
						  marlin_marshal_VOID__UINT64,
						  G_TYPE_NONE,
						  1, G_TYPE_UINT64);
}

/**
 * How the process functions work.
 * There are two threads: Thread A is the main GUI thread, thread B is the 
 * JACK process thread. Thread B is not able to do anything that is not
 * realtime safe. This includes the necessary things to access the frame
 * data, which is mmapped from disk and may require mmap calls and locking.
 *
 * When start_playing is called, an idle function is started that calls 
 * process_writer. This is run in thread A and is able to do all the mmapping
 * safely. Because it is run in the main thread, the frame data can be accessed
 * without locking. process_writer copies enough frame data to fill the ring
 * buffer. Once the ring buffer is filled, thread B is allowed to process the
 * data by reading from the ring buffer. As thread B removes data from the ring
 * buffer, thread A fills it back up.
 */

static void
process_playing (MarlinJack    *jack,
		 jack_nframes_t nframes)
{
	MarlinJackPrivate *priv;
	gboolean eos = FALSE;
	guint frames_needed = 0;
	guint64 tmp_pos;
	int i;
	
	priv = jack->priv;
	
	/* Always output in stereo, even when we've got a mono sample */
	for (i = 0; i < 2; i++) {
		struct _port_data *pd = priv->port_data[i];
		jack_default_audio_sample_t *o_buf;
		size_t bytes_avail;
		guint frames_avail, blank;
		
		bytes_avail = jack_ringbuffer_read_space (pd->rb);
		frames_avail = bytes_avail / frame_size;
		frames_needed = MIN (frames_avail, nframes);

		if (frames_needed == 0 && pd->writer_finished) {
			eos = TRUE;
		}

		o_buf = jack_port_get_buffer (pd->playback, nframes);
		jack_ringbuffer_read (pd->rb, (char *) o_buf,
				      frames_needed * frame_size);

		/* How many blank frames are needed at the end? */
		blank = nframes - frames_needed;
		if (blank > 0) {
			/* Need to zero the end of the buffer */
			memset (o_buf + frames_needed, 0, blank * frame_size);
		}
	}

	/* Fun aside: because the frames_needed is at Jack's sample rate
	   and priv->position needs them at the sample's sample rate
	   we need to estimate here */
	tmp_pos = priv->position;
	tmp_pos += (frames_needed / priv->src_ratio);
	if (tmp_pos > priv->finish) {
		tmp_pos = (tmp_pos - priv->finish) + priv->start;
	}
	priv->position = tmp_pos;
	
	priv->eos = eos;
}

static void
process_silence (MarlinJack    *jack,
		 jack_nframes_t nframes)
{
	MarlinJackPrivate *priv = jack->priv;
	int i;

	/* Always output in stereo even when we've got mono data */
	for (i = 0; i < 2; i++) {
		struct _port_data *pd = priv->port_data[i];
		jack_default_audio_sample_t *o_buf;
		
		o_buf = jack_port_get_buffer (pd->playback, nframes);
		/* Zero this so that there is silence */
		memset (o_buf, 0, 
			nframes * frame_size);
	}
}

static int
process_reader (jack_nframes_t nframes,
		void          *arg)
{
	MarlinJack *jack = (MarlinJack *) arg;
	MarlinJackPrivate *priv = jack->priv;

	switch (priv->mode) {
	case MARLIN_JACK_MODE_PLAYING:
		process_playing (jack, nframes);
		break;
		
	case MARLIN_JACK_MODE_RECORDING:
		break;

	case MARLIN_JACK_MODE_PREROLLING:
	case MARLIN_JACK_MODE_STOPPED:		
		process_silence (jack, nframes);
		break;

	case MARLIN_JACK_MODE_NOTHING:
	default:
		break;
	}

	return 0;
}

/* FIXME: Replace this with marlin_block_get_buffer */
static long
process_more_frames (gpointer userdata,
		     float  **data)
{
	struct _port_data *pd = userdata;
	MarlinJack *jack = pd->jack;
	MarlinJackPrivate *priv = jack->priv;
	guint frames_needed, copied;

	if (priv->repeating) {
		frames_needed = DEFAULT_RB_SIZE;
	} else {
		frames_needed = MIN (priv->finish - pd->offset,
				     DEFAULT_RB_SIZE);
	}

	if (frames_needed == 0) {
		/* This will tell the reader that we've finished writing */
		pd->writer_finished = TRUE;
		return 0;
	}

	copied = marlin_block_get_buffer (pd->current_block, pd->tmp_buffer,
					  pd->offset, frames_needed, 
					  &pd->current_block);

	pd->offset += copied;
	pd->offset_in_block = pd->offset - pd->current_block->start;
	
	*data = pd->tmp_buffer;

	return copied;
}

static gboolean
process_writer (gpointer data)
{
	MarlinJack *jack = data;
	MarlinJackPrivate *priv = jack->priv;
	int i;

	if (priv->mode != MARLIN_JACK_MODE_PLAYING &&
	    priv->mode != MARLIN_JACK_MODE_PREROLLING) {
		return TRUE;
	}

	if (priv->eos == TRUE) {
		marlin_jack_stop (jack);
		g_signal_emit (jack, signals[EOS], 0);
		return FALSE;
	}

 	g_signal_emit (jack, signals[POSITION_CHANGED], 0, priv->position);

	/* Always output in stereo even when we've got mono data */
	for (i = 0; i < 2; i++) {
		struct _port_data *pd = priv->port_data[i];
		size_t bytes_free;
		guint frames_needed;
		long generated;

		bytes_free = jack_ringbuffer_write_space (pd->rb);
#if 0
		/* Only fill it up when its half empty */
		if (bytes_free < (DEFAULT_RB_SIZE * frame_size) / 2) {
			continue;
		}
#endif
		frames_needed = bytes_free / frame_size;
		if (frames_needed == 0) {
			continue;
		}

		generated = src_callback_read (pd->state, priv->src_ratio,
					       frames_needed, pd->rb_buffer);
		if (generated == 0) {
			continue;
		}

		jack_ringbuffer_write (pd->rb, (char *) pd->rb_buffer,
				       generated * frame_size);
	}

	priv->mode = MARLIN_JACK_MODE_PLAYING;

	return TRUE;
}
	
static void
init (MarlinJack *jack)
{
	MarlinJackPrivate *priv;

	jack->priv = MARLIN_JACK_GET_PRIVATE (jack);
	priv = jack->priv;

	priv->mode = MARLIN_JACK_MODE_NOTHING;

	priv->eos = FALSE;
	priv->position = 0;
	priv->start = 0;
	priv->finish = 0;
	priv->repeating = FALSE;
}

GType
marlin_jack_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinJackClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinJack), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "MarlinJack",
					       &info, 0);
	}

	return type;
}

static int
srate_changed (jack_nframes_t nframes, 
	       gpointer       data)
{
	MarlinJack *jack = data;
	MarlinJackPrivate *priv = jack->priv;

	priv->jrate = nframes;
	priv->src_ratio = calculate_rate_ratio (priv->rate, priv->jrate);
	
	return 0;
}

static int
bufsize_changed (jack_nframes_t nframes,
		 gpointer       data)
{
	MarlinJack *jack = data;
	MarlinJackPrivate *priv = jack->priv;

	priv->bufsize = nframes;

	g_free (priv->jack_buf);
	priv->jack_buf = g_new0 (jack_default_audio_sample_t, nframes * 16);

	return 0;
}

/**
 * marlin_jack_new:
 * @client_name: Name of this client
 * @server_name: Name of the jack server to connect to (or NULL for default)
 * @error:
 *
 * Creates a new #MarlinJack structure that will be connected to @server_name
 * (or the default server if @server_name is NULL).
 */
MarlinJack *
marlin_jack_new (const char *client_name,
		 const char *server_name,
		 GError    **error)
{
	MarlinJack *jack;
	MarlinJackPrivate *priv;
	jack_options_t options = JackNullOption;
	jack_status_t status;

	jack = g_object_new (MARLIN_JACK_TYPE, NULL);
	priv = jack->priv;
	
	if (server_name) {
		options |= JackServerName;
	}

	priv->client = jack_client_open (client_name, options,
					 &status, server_name);
	if (priv->client == NULL) {
		/* Make error */
		g_warning ("No jacky client\n");
		g_object_unref (jack);
		return NULL;
	}

	jack_set_process_callback (priv->client, process_reader, jack);
	jack_set_buffer_size_callback (priv->client, bufsize_changed, jack);
	priv->bufsize = jack_get_buffer_size (priv->client);
	priv->jack_buf = g_new0 (jack_default_audio_sample_t, priv->bufsize * 16);

	jack_set_sample_rate_callback (priv->client, srate_changed, jack);
	priv->jrate = jack_get_sample_rate (priv->client);

	return jack;
}

gboolean
marlin_jack_start_playback (MarlinJack *jack,
			    guint64     start,
			    guint64     finish,
			    gboolean    repeating,
			    GError    **error)
{
	MarlinJackPrivate *priv;
	const char **ports;
	int i;

	priv = jack->priv;

	if (priv->sample == NULL) {
		return FALSE;
	}

	if (jack_activate (priv->client)) {
		return FALSE;
	}

	marlin_sample_read_lock (priv->sample);

	priv->start = start;
	priv->finish = finish;
	priv->repeating = repeating;

	ports = jack_get_ports (priv->client, NULL, NULL,
				JackPortIsPhysical | JackPortIsInput);
	/* Even if we've got mono data, we output in stereo */
	for (i = 0; i < 2; i++) {
		struct _port_data *pd = priv->port_data[i];
		int err;

		if (ports[i] == NULL) {
			marlin_sample_read_unlock (priv->sample);
			return FALSE;
		}

		if (jack_connect (priv->client,
				  jack_port_name (pd->playback),
				  ports[i])) {
			marlin_sample_read_unlock (priv->sample);
			return FALSE;
		}
		
		if (i < priv->channels) {
			pd->channel = marlin_sample_get_channel (priv->sample, 
								 i);
		} else {
			pd->channel = marlin_sample_get_channel (priv->sample, 0);
		}
		pd->current_block = marlin_channel_get_block_for_frame (pd->channel, start);
		pd->offset = start;
		pd->offset_in_block = pd->offset - pd->current_block->start;

		pd->writer_finished = FALSE;

		/* Create the ring buffer */
		pd->rb = jack_ringbuffer_create (frame_size * DEFAULT_RB_SIZE);

		pd->rb_buffer = g_new (float, DEFAULT_RB_SIZE);
		pd->tmp_buffer = g_new (float, DEFAULT_RB_SIZE);

		pd->state = src_callback_new (process_more_frames, 
					      SRC_SINC_BEST_QUALITY, 
					      1, &err, pd);
	}
	free (ports);

	priv->eos = FALSE;
	priv->position = start;
	priv->mode = MARLIN_JACK_MODE_PREROLLING;

	priv->writer_id = g_idle_add (process_writer, jack);

	return TRUE;
}

gboolean
marlin_jack_start_recording (MarlinJack *jack,
			     GError    **error)
{
	return TRUE;
}

void
marlin_jack_pause (MarlinJack *jack)
{
	MarlinJackPrivate *priv;

	priv = jack->priv;

	priv->mode = MARLIN_JACK_MODE_STOPPED;
}

void
marlin_jack_continue (MarlinJack *jack)
{
	MarlinJackPrivate *priv;

	priv = jack->priv;

	priv->mode = MARLIN_JACK_MODE_PLAYING;
}

void
marlin_jack_stop (MarlinJack *jack)
{
	MarlinJackPrivate *priv;
	const char **ports;
	int i;

	priv = jack->priv;

	priv->mode = MARLIN_JACK_MODE_STOPPED;

	jack_deactivate (priv->client);

	marlin_sample_read_unlock (priv->sample);

	g_source_remove (priv->writer_id);
	priv->writer_id = 0;

	ports = jack_get_ports (priv->client, NULL, NULL,
				JackPortIsPhysical | JackPortIsInput);
	for (i = 0; i < priv->channels; i++) {
		struct _port_data *pd = priv->port_data[i];

		jack_disconnect (priv->client,
				 jack_port_name (pd->playback), ports[i]);
		pd->channel = NULL;
		pd->current_block = NULL;

		if (pd->rb) {
			jack_ringbuffer_free (pd->rb);
			g_free (pd->rb_buffer);
			g_free (pd->tmp_buffer);

			pd->rb = NULL;
			pd->rb_buffer = NULL;
			pd->tmp_buffer = NULL;
		}

		src_delete (pd->state);

/* 		src_reset (pd->state); */
	}
	free (ports);
}

