/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@gnome.org>
 *
 *  Copyright 2003 - 2006 Iain Holmes
 *
 *  This file is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU Library General Public
 *  License as published by the Free Software Foundation;
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *
 */

#include <glib/gi18n.h>

#include <marlin/marlin-program.h>
#include <marlin/marlin-marshal.h>

enum {
	NEW_VIEW,
	CLIPBOARD_CHANGED,
	LAST_SIGNAL
};

static GObjectClass *parent_class = NULL;
static MarlinProgram *main_program = NULL;
static guint signals[LAST_SIGNAL];

struct _MarlinProgramPrivate {
	MarlinSample *clipboard;
	char *clipboard_src;
};

static void
finalize (GObject *object)
{
	MarlinProgram *program;
	MarlinProgramPrivate *priv;

	program = MARLIN_PROGRAM (object);
	priv = program->priv;

	g_free (priv->clipboard_src);

	g_free (priv);
	program->priv = NULL;

	parent_class->finalize (object);
}

static void
dispose (GObject *object)
{
	MarlinProgram *program;
	MarlinProgramPrivate *priv;
	
	program = MARLIN_PROGRAM (object);
	priv = program->priv;

	if (priv->clipboard != NULL) {
 		g_object_unref (G_OBJECT (priv->clipboard));
		priv->clipboard = NULL;
	}

	parent_class->dispose (object);
}

static void
class_init (MarlinProgramClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;

	object_class->finalize = finalize;
	object_class->dispose = dispose;

	parent_class = g_type_class_peek_parent (klass);

	signals[NEW_VIEW] = g_signal_new ("new-view",
					  G_TYPE_FROM_CLASS (klass),
					  G_SIGNAL_RUN_FIRST |
					  G_SIGNAL_NO_RECURSE,
					  G_STRUCT_OFFSET (MarlinProgramClass, new_view),
					  NULL, NULL,
					  marlin_marshal_VOID__OBJECT,
					  G_TYPE_NONE, 1,
					  G_TYPE_OBJECT);
	signals[CLIPBOARD_CHANGED] = g_signal_new ("clipboard-changed",
						   G_TYPE_FROM_CLASS (klass),
						   G_SIGNAL_RUN_FIRST |
						   G_SIGNAL_NO_RECURSE,
						   G_STRUCT_OFFSET (MarlinProgramClass, clipboard_changed),
						   NULL, NULL,
						   g_cclosure_marshal_VOID__VOID,
						   G_TYPE_NONE, 0);
}

#if 0
static void
pipeline_eos (GstElement *element,
	      MarlinProgram *program)
{
	marlin_pipeline_set_state (MARLIN_PIPELINE (program->priv->pipeline), 
				   GST_STATE_READY);
}
#endif

static void
init (MarlinProgram *program)
{
	program->priv = g_new0 (MarlinProgramPrivate, 1);
#if 0
	program->priv->pipeline = g_object_new (MARLIN_PLAY_PIPELINE_TYPE, NULL);
	g_signal_connect (program->priv->pipeline, "eos",
			  G_CALLBACK (pipeline_eos), program);
#endif
}

GType
marlin_program_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinProgramClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinProgram), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "MarlinProgram", &info, 0);
	}

	return type;
}

/**
 * marlin_program_get_default:
 * 
 * Obtains the default MarlinProgram for this instance of Marlin.
 *
 * Return value: A pointer to MarlinProgram.
 */
MarlinProgram *
marlin_program_get_default (void)
{
	if (main_program != NULL) {
		return main_program;
	}

	main_program = g_object_new (MARLIN_PROGRAM_TYPE, NULL);
	return main_program;
}

/**
 * marlin_program_destroy_default:
 *
 * Frees the default program. Should only be called by Marlin.
 */
void
marlin_program_destroy_default (void)
{
	if (main_program == NULL) {
		return;
	}

	/* Should this be held in a mutex? */
	g_object_unref (main_program);
	main_program = NULL;
}

/**
 * marlin_program_get_clipboard:
 * @program:
 *
 * Obtains the clipboard associated with @program.
 *
 * Returns: A #MarlinSample.
 */
MarlinSample *
marlin_program_get_clipboard (MarlinProgram *program)
{
	return (MarlinSample *) program->priv->clipboard;
}

/**
 * marlin_program_set_clipboard:
 * @program:
 * @clipboard_object:
 * @src:
 *
 * Sets the clipboard of @program to @clipboard_object and sets the
 * source description to @src.
 */
void
marlin_program_set_clipboard (MarlinProgram *program,
			      MarlinSample *clipboard_object,
			      const char *src)
{
	if (program->priv->clipboard != NULL) {
		g_object_unref (program->priv->clipboard);
	}

	program->priv->clipboard = (MarlinSample *) clipboard_object;
	if (program->priv->clipboard != NULL) {
		g_object_ref (program->priv->clipboard);
	}

	g_object_set (clipboard_object,
		      "name", _("Clipboard"),
		      NULL);

	if (program->priv->clipboard_src != NULL) {
		g_free (program->priv->clipboard_src);
	}

	program->priv->clipboard_src = g_strdup (src);

	g_signal_emit (G_OBJECT (program), signals[CLIPBOARD_CHANGED], 0);
}

/**
 * marlin_program_get_clip_source:
 * @program:
 * 
 * Obtains the source of the clipboard.
 * FIXME: Should this be combined with get_clipboard?
 */
const char *
marlin_program_get_clip_source (MarlinProgram *program)
{
	return program->priv->clipboard_src;
}

/* FIXME: Should these be moved into the clipboard-info dialog? */
/**
 * marlin_program_play_clipboard:
 * @program:
 * 
 * Starts the clipboard playing.
 */
void
marlin_program_play_clipboard (MarlinProgram *program)
{
#if 0
	marlin_pipeline_set_state (MARLIN_PIPELINE (program->priv->pipeline),
				   GST_STATE_PLAYING);
#endif
}

/** 
 * marlin_program_stop_clipboard:
 * @program:
 *
 * Stop the clipboard playing.
 */
void
marlin_program_stop_clipboard (MarlinProgram *program)
{
#if 0
	marlin_pipeline_set_state (MARLIN_PIPELINE (program->priv->pipeline),
				   GST_STATE_READY);
#endif
}

/**
 * marlin_program_request_new_view:
 * @program: A #MarlinProgram.
 * @sample: A #MarlinSample.
 * 
 * Requests that a new view be created for @sample.
 */
void
marlin_program_request_new_view (MarlinProgram *program,
				 MarlinSample *sample)
{
	g_signal_emit (G_OBJECT (program), signals[NEW_VIEW], 0, sample);
}
